import inspect

import wrapt
from nameko.extensions import DependencyProvider

from .commands import CommandsWrapper, CommandsProxy


class _DispatchCommands(CommandsWrapper):
    pass


class DispatchCommandsDependencyProvider(DependencyProvider):

    """Dependency provider to access to the `::DispatchCommands::`.

    It will return a ``DispatchCommandsWrapper`` instance.
    """

    @classmethod
    def dispatch_after_commit(cls, wrapped=None, *args, **kwargs):
        """Execute the dispatch event when the function call succeeds

        The following example demonstrates the use of::

            >>> from nameko.rpc import rpc
            >>> from nameko.events import EventDispatcher
            >>> from nameko_atomicity import DispatchCommands
            ...
            >>> class ConversionService(object):
            ...    name = "conversions"
            ...    dispatch_commands = DispatchCommands()
            ...
            ...    @rpc
            ...    @dispatch_after_commit
            ...    def inches_to_cm(self, inches):
            ...        event_name = "booking_updated"
            ...        dispatch_data = {}
            ...        self.dispatch_commands.append(
            ...            func=self.event_dispatcher,
            ...            args=(event_name, dispatch_data),
            ...            kwargs={},
            ...        )

        """

        def find_commands_providers(instance):
            commands_provides = inspect.getmembers(
                instance, lambda obj: isinstance(obj, _DispatchCommands)
            )
            return commands_provides

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            commands_provides = find_commands_providers(instance)
            commands_provides = [
                commands_provide[1] for commands_provide in commands_provides
            ]
            try:
                response = wrapped(*args, **kwargs)
                CommandsProxy(commands_provides).exec_commands()
                return response

            finally:
                CommandsProxy(commands_provides).clear_commands()

        if wrapped:
            return wrapper(wrapped)

        return wrapper

    def get_dependency(self, worker_ctx):
        return _DispatchCommands(worker_ctx)


DispatchCommands = DispatchCommandsDependencyProvider
dispatch_after_commit = DispatchCommandsDependencyProvider.dispatch_after_commit



from collections import namedtuple, UserList
from typing import Union

from .exceptions import InvalidCommand

Command = namedtuple("Command", "func, args, kwargs")
""" Define a command class that must contain three parameters: func, args, kwargs """


class Commands(UserList):
    """The base class of Commands"""

    def append_command(self, command: Command):
        self.validate(command)
        super(Commands, self).append(command)

    def insert_command(self, index: int, command: Command):
        self.validate(command)
        super(Commands, self).insert(index, command)

    @staticmethod
    def validate(command):
        if not isinstance(command, Command):
            raise TypeError(
                f"Must be an instance of `:Command:`, "
                f"but now it's `{command.__class__}`."
            )

    @staticmethod
    def _adapt_parameters(args, kwargs):
        args = tuple() if not args else args
        kwargs = dict() if not kwargs else kwargs
        return args, kwargs

    def append(
        self, func, args: Union[tuple, list] = None, kwargs: dict = None
    ) -> None:
        args, kwargs = self._adapt_parameters(args, kwargs)

        command = Command(func=func, args=args, kwargs=kwargs)
        self.append_command(command)

    def insert(
        self,
        index,
        func,
        args: Union[tuple, list] = None,
        kwargs: dict = None,
    ) -> None:
        args, kwargs = self._adapt_parameters(args, kwargs)

        command = Command(func=func, args=args, kwargs=kwargs)
        self.insert_command(index, command)

    def exec_commands(self) -> list:
        commands = self.copy()
        self.clear()

        for command in commands:
            try:
                command.func(*command.args, **command.kwargs)
            except Exception as e:
                raise InvalidCommand({"command": command, "error_message": e})

        return list(commands)

    def clear_commands(self) -> None:
        self.clear()


class CommandsWrapper(object):
    def __init__(self, worker_ctx, commands_cls=Commands):
        self.worker_ctx = worker_ctx
        self._commands = commands_cls()

    def append_command(self, command: Command):
        self._commands.append_command(command)

    def insert_command(self, index: int, command: Command):
        self._commands.insert_command(index, command)

    def append(
        self, func, args: Union[tuple, list] = None, kwargs: dict = None
    ) -> None:
        self._commands.append(func, args, kwargs)

    def insert(
        self,
        index,
        func,
        args: Union[tuple, list] = None,
        kwargs: dict = None,
    ) -> None:
        self._commands.insert(index, func, args, kwargs)

    def exec_commands(self) -> list:
        return self._commands.exec_commands()

    def clear_commands(self):
        return self._commands.clear_commands()


class CommandsProxy(object):
    def __init__(self, items):
        self._items = items

    def __getattr__(self, name):
        def spawning_method(*args, **kwargs):
            items = self._items

            def call(item):
                return getattr(item, name)(*args, **kwargs)

            if items:
                return list(map(call, self._items))

        return spawning_method



from .dispatch_commands import DispatchCommands  # noqa
from .dispatch_commands import dispatch_after_commit  # noqa
from .rollback_commands import RollbackCommands  # noqa
from .rollback_commands import rollback_once_failed  # noqa



import eventlet
from eventlet import Event

import pytest
from mock import Mock
from nameko.rpc import rpc
from nameko.testing.services import entrypoint_hook

from nameko_atomicity.rollback_commands import (
    RollbackCommands,
    rollback_once_failed,
)


class TestRollbackCommands:
    @pytest.fixture
    def tracker(self):
        return Mock()

    @pytest.fixture(params=["dependency", "dependency_proxy"])
    def service_class(self, request, tracker):
        class ExampleService:
            name = "exampleservice"

            rollback_commands = RollbackCommands()

            @rollback_once_failed
            @rpc
            def calculate(self, data):
                self.rollback_commands.append(
                    func=self._calculate, args=(data,), kwargs={}
                )
                data["release"].wait()
                raise Exception("Boom!")

            @rollback_once_failed
            @rpc
            def calculate_no_raise(self, data):
                self.rollback_commands.append(
                    func=self._calculate, args=(data,), kwargs={}
                )
                data["release"].wait()

            def _calculate(self, data):
                tracker(data)

        class ExampleServiceUsingProxyDecorator:
            name = "exampleservice"

            rollback_commands = RollbackCommands()

            @rollback_once_failed()
            @rpc
            def calculate(self, data):
                self.rollback_commands.append(
                    func=self._calculate, args=(data,), kwargs={}
                )
                data["release"].wait()
                raise Exception("Boom!")

            @rollback_once_failed()
            @rpc
            def calculate_no_raise(self, data):
                self.rollback_commands.append(
                    func=self._calculate, args=(data,), kwargs={}
                )
                data["release"].wait()

            def _calculate(self, data):
                tracker(data)

        services = {
            "dependency": ExampleService,
            "dependency_proxy": ExampleServiceUsingProxyDecorator,
        }

        return services[request.param]

    def test_rollback_after_commit_end_to_end(
        self, rabbit_config, container_factory, service_class, tracker
    ):
        release_one, release_two = Event(), Event()

        container = container_factory(service_class, rabbit_config)
        container.start()

        def make_coroutine(release):
            def coroutine():
                with entrypoint_hook(container, "calculate") as calculate:
                    data = {"id": 222, "release": release}
                    calculate(data)

            return coroutine

        def make_coroutine_no_raise(release):
            def coroutine():
                with entrypoint_hook(
                    container, "calculate_no_raise"
                ) as calculate_no_raise:
                    data = {"id": 222, "release": release}
                    calculate_no_raise(data)

            return coroutine

        coroutine_one = make_coroutine(release_one)
        assert 0 == tracker.call_count

        thread_one = eventlet.spawn(coroutine_one)
        eventlet.sleep(0.1)

        release_one.send()
        pytest.raises(Exception, thread_one.wait)
        assert 1 == tracker.call_count

        # no raise
        coroutine_two = make_coroutine_no_raise(release_two)
        assert 1 == tracker.call_count

        thread_two = eventlet.spawn(coroutine_two)
        eventlet.sleep(0.1)

        release_two.send()
        thread_two.wait()
        assert 1 == tracker.call_count



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name='nameko-atomicity',
    version='1.0.0',
    description='Atomicity dependency for nameko services',
    author='jeremy-jin',
    author_email='306795597@qq.com',
    url='https://github.com/jeremy-jin/nameko-atomicity',
    packages=find_packages(exclude=("test",)),
    install_requires=[
        "nameko>=2.0.0",
    ],
    extras_require={
        'dev': [
            "coverage==5.5",
            "flake8==3.9.2",
            "pylint==2.8.3",
            "pytest==6.2.4",
            "black==21.5b2"
        ]
    },
    zip_safe=True,
    license='Apache License, Version 2.0',
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ]
)


import pytest

from nameko_atomicity.exceptions import InvalidCommand
from nameko_atomicity.commands import (
    CommandsWrapper,
    Command,
    Commands,
)


class Container(object):
    pass


class WorkerCtx:
    container = Container()


def func(*args, **kwargs):
    pass


class TestCommands:
    def func(*args, **kwargs):
        pass

    def func1(*args, **kwargs):
        raise TypeError("Error")

    class Container(object):
        pass

    commands = Commands()

    def test_append_command(self):
        self.commands.clear()
        command = Command(func=self.func, args=(), kwargs={})
        self.commands.append_command(command)
        assert self.commands
        assert len(self.commands) == 1
        assert self.commands[0] is command

    def test_append_command_with_type_error(self):
        pytest.raises(TypeError, self.commands.append_command, {})

    def test_insert_command(self):
        self.commands.clear()
        command = Command(func=self.func, args=(), kwargs={})
        command1 = Command(func=self.func, args=(), kwargs={})
        self.commands.insert_command(1, command)
        self.commands.insert_command(0, command1)

        assert self.commands
        assert len(self.commands) == 2
        assert self.commands[0] is command1
        assert self.commands[1] is command

    def test_append(self):
        self.commands.clear()
        self.commands.append(self.func)

        assert self.commands
        assert len(self.commands) == 1

    def test_insert(self):
        self.commands.clear()
        self.commands.insert(0, self.func)

        assert self.commands
        assert len(self.commands) == 1

    def test_exec_commands(self):
        self.commands.clear()
        self.commands.append(self.func)
        commands = self.commands.exec_commands()

        assert commands
        assert len(commands) == 1
        assert self.commands == []

    def test_exec_commands_with_invalid_command_error(self):
        self.commands.clear()
        self.commands.append(self.func1)
        pytest.raises(InvalidCommand, self.commands.exec_commands)

    def test_clear_commands(self):
        self.commands.clear()
        self.commands.append(self.func)

        assert self.commands
        assert len(self.commands) == 1

        self.commands.clear_commands()

        assert self.commands == []


class TestCommandsWrapper:
    worker_ctx = WorkerCtx()
    command_wrapper = CommandsWrapper(worker_ctx)
    command = Command(func=func, args=(), kwargs={})

    def test_add_command(self):
        self.command_wrapper.append_command(self.command)
        assert self.command_wrapper._commands
        assert len(self.command_wrapper._commands) == 1
        self.command_wrapper.clear_commands()
        assert len(self.command_wrapper._commands) == 0

    def test_insert_command(self):
        self.command_wrapper.clear_commands()
        self.command_wrapper.insert_command(0, self.command)
        command1 = Command(func=func, args=(), kwargs={})
        self.command_wrapper.insert_command(0, command1)

        assert self.command_wrapper._commands
        assert len(self.command_wrapper._commands) == 2
        assert self.command_wrapper._commands[0] == command1

    def test_append(self):
        self.command_wrapper.clear_commands()
        self.command_wrapper.append(func)
        assert self.command_wrapper._commands
        assert len(self.command_wrapper._commands) == 1

    def test_insert(self):
        self.command_wrapper.clear_commands()
        self.command_wrapper.insert(0, func)
        assert self.command_wrapper._commands
        assert len(self.command_wrapper._commands) == 1

    def test_exec_commands(self):
        self.command_wrapper.clear_commands()
        self.command_wrapper.insert(0, func)
        commands = self.command_wrapper.exec_commands()
        assert commands
        assert len(commands) == 1



import inspect

import wrapt
from nameko.extensions import DependencyProvider

from .commands import CommandsWrapper, CommandsProxy


class _RollbackCommands(CommandsWrapper):
    pass


class RollbackCommandsDependencyProvider(DependencyProvider):

    """Dependency provider to access to the `::RollbackCommands::`.

    It will return a ``RollbackCommandsWrapper`` instance.
    """

    @classmethod
    def rollback_once_failed(cls, wrapped=None):
        """Execute the rollback command when a function call fails

        The following example demonstrates the use of::

            >>> from nameko.rpc import rpc
            ... from nameko_atomicity import RollbackCommands
            ...
            ... def rollback_function():
            ...     pass
            ...
            ... class ConversionService(object):
            ...    name = "conversions"
            ...
            ...    rollback_commands = RollbackCommands()
            ...
            ...    @rpc
            ...    @rollback_once_failed
            ...    def inches_to_cm(self, inches):
            ...        self.rollback_commands.append(
            ...            func=rollback_function,
            ...            args=(),
            ...            kwargs={},
            ...        )

        """

        def find_commands_providers(instance):
            commands_provides = inspect.getmembers(
                instance, lambda obj: isinstance(obj, _RollbackCommands)
            )
            return commands_provides

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            commands_provides = find_commands_providers(instance)
            commands_provides = [
                commands_provide[1] for commands_provide in commands_provides
            ]

            try:
                response = wrapped(*args, **kwargs)
            except Exception as exc:
                CommandsProxy(commands_provides).exec_commands()

                raise exc

            finally:
                CommandsProxy(commands_provides).clear_commands()

            return response

        if wrapped:
            return wrapper(wrapped)

        return wrapper

    def get_dependency(self, worker_ctx):
        return _RollbackCommands(worker_ctx)


RollbackCommands = RollbackCommandsDependencyProvider
rollback_once_failed = RollbackCommandsDependencyProvider.rollback_once_failed
